// EPCGP suite, 1D condensate, no spin effects
// ===========================================
//
// The set of two differential equations (Gross-Pitaevskii), with
// nonlinearities and with decay.  Program computes initial state and
// then uses it for solving the equation.  Every print_tstep it prints
// out the amplitudes of psi_x and psi_c and their Fourier transforms.
//
// Parameters are set in the source code in order to ensure best
// optimization by the compiler.

#include <iostream>
#include <iomanip>
#include <cmath>
#include <complex>

using namespace std;

typedef complex<double> cdouble; 

// Helper constants.
const cdouble I = cdouble(0, 1); // Imaginary unit.
const double pi = 4 * atan(1);   // π.
const double hbar = 0.6582;      // Reduced Planck constant [meV·ps].

// Parameters of the setup.
const double d = 5;    // Width of the 1D condensate [µm].
const double Fp = 0.5; // Amplitude of the pump field [meV/(µm)^(1/2)].
const double kp = 0;   // Momentum of the pump field [1/(µm)].
const double deltaomega = 0; // Detuning of the pump field [1/ps].
const double w = 10;     // Spread of the pump [µm].
const double delta = 0;  // Detuning of polaritons [meV].
const double g_2D = 0.01; // Material parameter for 2D case [meV·(µm)²].
const double g = g_2D / (sqrt(2 * pi) * d); // Material parameter [meV·µm]
const double OmegaR = 4.4;  // Rabi frequancy [meV].
const double gammax = 0.01; // Decay rate of an exciton [1/(ps)].
const double gammac = 0.1;  // Decay rate of a photon [1/(ps)].
const double cavsizex = 100;    // Cavity size [µm].
const double m0 = 5.677e3;      // Mass of free electron [meV/(µm/ps)²].
const double mc = m0 * 2e-5;    // Effective mass of polaritons [meV].
const double h = 0.001;         // Time step [ps].

// Parameters of the simulation.
const int xsize = 201;         // Number of mesh nodes along X axis.
const double sx = cavsizex / (xsize - 1); // Spatial step along X axis.
const double xmin = -(cavsizex / 2); // Lower boundary of X coordinate.
const double xmax = (cavsizex / 2); // Upper boundary of X coordinate.
const int nsteps = 2000;            // Number of simulation iterations.
const int print_tstep = 10;          // Print every nth time step.
const int print_xstep = 1;      // Print every nth spatial point.

// Global variales.
cdouble psix[xsize];            // Array for values of psix [1/(µm)^(1/2)].
cdouble psic[xsize];            // Array for values of psic [1/(µm)^(1/2)].
double t;                       // Current time [ps].

// Computes square of complex x.
inline cdouble square (cdouble x)
{
    return x * x;
}

// Computes square of real x.
inline double square (double x) 
{
    return x * x;
}

// Computes F.
cdouble F(double x, double t)
{
    return Fp * exp(I * (kp * x - deltaomega * t))
        * exp(-square(x) / (2 * square(w)));
}

// Initializes psix, psix and t.
void initialization()
{
    t = 0;
    for (int i = 0; i < xsize; ++i) {
        psix[i] = 0;
        psic[i] = 0;
    }
}

// Computes single step for psi_x: f(psix, psic, t).  The results are
// returned in a freshly allocated array.
cdouble *compute_psix_substep(cdouble psix[], cdouble psic[], double t)
{
    cdouble *k = new cdouble[xsize];
    #pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        k[i] = (-I * OmegaR * psic[i] / hbar
                + ((-gammax * 0.5 - I * g * square(abs(psix[i])) / hbar
                    - I * delta / hbar)
                   * psix[i]));
    }
    return k;
}

// Computes derivative  f(psic, psix, t).  The results are
// returned in a freshly allocated array.
cdouble *compute_psic_substep(cdouble psic[], cdouble psix[], double t)
{
    cdouble *k = new cdouble[xsize];
    #pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        const double x = i * sx + xmin;
        const cdouble psicm1 = i > 0 ? psic[i - 1] : 0;
        const cdouble psicp1 = i < xsize - 1 ? psic[i + 1] : 0;
        k[i] = (-I * F(x, t) / hbar - I * OmegaR * psix[i] / hbar
                - gammac * 0.5 * psic[i]
                + (I * hbar / (2 * mc) / square(sx)
                   * (psicm1 - 2.0 * psic[i] + psicp1)));
    }
    return k;
}

// Computes x+y*c for scalar coefficient c and each pair of the
// elements of arrays x and y separately.  The results are returned in
// a freshly allocated array.
cdouble *increment_substep(cdouble x[], cdouble y[], double c)
{
    cdouble *k = new cdouble[xsize];
    #pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        k[i] = x[i] + y[i] * c;
    }
    return k;
}

// Computes a single time step for psix and psic simultaneously.
void compute_step()
{
    // Substep 1.  Compute set of k1x and k1c coefficients starting
    // from psix and psic and time t.
    cdouble *k1x = compute_psix_substep(psix, psic, t);
    cdouble *k1c = compute_psic_substep(psic, psix, t);

    // Substep 2a. Compute values of psix and psic at time t + h / 2
    // using Euler's method.
    cdouble *psix2 = increment_substep(psix, k1x, h * 0.5);
    cdouble *psic2 = increment_substep(psic, k1c, h * 0.5);

    // Substep 2b. Compute set of k2x and k2c coefficients.
    cdouble *k2x = compute_psix_substep(psix2, psic2, t + h * 0.5);
    cdouble *k2c = compute_psic_substep(psic2, psix2, t + h * 0.5);

    // Substep 3a. Compute values of psix and psic at time t + h / 2
    // using Euler's method.
    cdouble *psix3 = increment_substep(psix, k2x, h * 0.5);
    cdouble *psic3 = increment_substep(psic, k2c, h * 0.5);

    // Substep 3b. Compute set of k3x and k3c coefficients.
    cdouble *k3x = compute_psix_substep(psix3, psic3, t + h * 0.5);
    cdouble *k3c = compute_psic_substep(psic3, psix3, t + h * 0.5);

    // Substep 4a. Compute values of psix and psic at time t.
    cdouble *psix4 = increment_substep(psix, k3x, h);
    cdouble *psic4 = increment_substep(psic, k3c, h);

    // Substep 4b. Compute set of k4x and k4c coefficients.
    cdouble *k4x = compute_psix_substep(psix4, psic4, t + h);
    cdouble *k4c = compute_psic_substep(psic4, psix4, t + h);

    // Substep 5. Compute new values of psix and psic for t + h.
    for (int i = 0; i < xsize; ++i) {
        psix[i] += (k1x[i] + 2.0 * k2x[i] + 2.0 * k3x[i] + k4x[i]) * (h / 6);
        psic[i] += (k1c[i] + 2.0 * k2c[i] + 2.0 * k3c[i] + k4c[i]) * (h / 6);
    }

    // Compute new t.
    t += h;

    // Free the memory.
    delete[] k1x;
    delete[] k1c;
    delete[] k2x;
    delete[] k2c;
    delete[] k3x;
    delete[] k3c;
    delete[] k4x;
    delete[] k4c;
    delete[] psix2;
    delete[] psic2;
    delete[] psix3;
    delete[] psic3;
    delete[] psix4;
    delete[] psic4;
}

int main(int argc, char* argv[])
{
    initialization();
    double maxv = 0;
    for (int n=0; n < nsteps; ++n) {
        double sum = 0;
        for (int i = 0; i < xsize; ++i) {
            const double px = square(abs(psix[i]));
            const double pc = square(abs(psic[i]));
            maxv = max(maxv, max(px, pc));
            sum += (pc + px) * sx;
        }
        if (n % print_tstep == 0) {
            cerr << "t = " << t << ", sum = " << sum << endl;
            cout << setprecision(4) << t;
            for (int i = 0; i < xsize; ++i) {
                cout << '\t' << setprecision(5) << norm(psic[i]);
            }
            for (int i = 0; i < xsize; ++i) {
                cout << '\t' << setprecision(5) << norm(psix[i]);
            }
	    cout << endl;
        }
        compute_step();
    }
    cerr << "maxv = " << maxv << endl;
}                             
