// EPCGP suite, 1D condensate, spin effects included
// =================================================
//
// The set of four differential equations (Gross-Pitaevskii), with
// nonlinearities and with decay.  Program computes initial state and
// then uses it for solving the equation.  Every print_tstep it prints
// out the amplitudes of psi_x and psi_c and their Fourier transforms.
//
// Parameters are set in the source code in order to ensure best
// optimization by the compiler.

#include <algorithm>
#include <iostream>
#include <iomanip>
#include <cmath>
#include <complex>
#include <ctime>

using namespace std;

typedef complex<double> cdouble; 

// Helper constants.
const cdouble I = cdouble(0, 1); // Imaginary unit.
const double pi = 4 * atan(1);   // π.
const double hbar = 0.6582;      // Reduced Planck constant [meV·ps].

// Parameters of the setup.
const double d = 5; // Width of the 1D condensate [µm].
const double FpA = 0.5; // Amplitude of the pump field, σ = +1 [meV/(µm)^(1/2)].
const double kpA = 0; // Momentum of the pump field, σ = +1 [1/(µm)].
const double deltaomegaA = 5; // Detuning of the pump field, σ = +1 [1/ps].
const double wA = 10; // Spread of the pump, σ = +1  [µm].
const double FpB = 0.5; // Amplitude of the pump field, σ = -1 [meV/(µm)^(1/2)].
const double kpB = 0; // Momentum of the pump field, σ = -1 [1/(µm)].
const double deltaomegaB = 5; // Detuning of the pump field, σ = -1 [1/ps].
const double wB = 10; // Spread of the pump, σ = -1 [µm].
const double delta = 0; // Detuning of polaritons [meV].
const double g1_2D = 50; // Material parameter 1 for 2D case [meV·(µm)²].
const double g1 = g1_2D / (sqrt(2 * pi) * d); // Material parameter 1 [meV·µm]
const double g2_2D = -100; // Material parameter 2 for 2D case [meV·(µm)²].
const double g2 = g2_2D / (sqrt(2 * pi) * d); // Material parameter 2 [meV·µm]
const double OmegaR = 4.4; // Rabi frequancy [meV].
const double gammax = 0.05; // Decay rate of an exciton [1/(ps)].
const double gammac = 0.5; // Decay rate of a photon [1/(ps)].
const double cavsizex = 100; // Cavity size [µm].
const double m0 = 5.677e3; // Mass of free electron [meV/(µm/ps)²].
const double mc = -m0 * 2e-5; // Effective mass of polaritons [meV].
const double h = 0.001; // Time step [ps].

// Parameters of the simulation.
const int xsize = 201;          // Number of mesh nodes along X axis.
const double sx = cavsizex / (xsize - 1); // Spatial step along X axis.
const double xmin = -(cavsizex / 2); // Lower boundary of X coordinate.
const double xmax = (cavsizex / 2); // Upper boundary of X coordinate.
const int nsteps = 2000;           // Number of simulation iterations.
const int print_tstep = 10;             // Print every nth time step.
const int print_xstep = 1;       // Print every nth spatial point.

// Global variales.
cdouble psixA[xsize]; // Array for values of psix, spin σ = +1 [1/(µm)^(1/2)].
cdouble psicA[xsize]; // Array for values of psic, spin σ = +1 [1/(µm)^(1/2)].
cdouble psixB[xsize]; // Array for values of psix, spin σ = -1 [1/(µm)^(1/2)].
cdouble psicB[xsize]; // Array for values of psic, spin σ = -1 [1/(µm)^(1/2)].
double t;             // Current time [ps].

// Computes square of complex x.
inline cdouble square (cdouble x)
{
    return x * x;
}

// Computes square of real x.
inline double square (double x) 
{
    return x * x;
}

// Computes F for spin σ = +1.
inline cdouble FA(double x, double t)
{
    return FpA * exp(I * (kpA * x - deltaomegaA * t))
        * exp(-square(x) / (2 * square(wA)));
}

// Computes F for spin σ = -1.
inline cdouble FB(double x, double t)
{
    return FpB * exp(I * (kpB * x - deltaomegaB * t))
        * exp(-square(x) / (2 * square(wB)));
}

// Initializes psix, psix and t.
void initialization()
{
    t = 0;
    for (int i = 0; i < xsize; ++i) {
        psixA[i] = 0;
        psicA[i] = 0;
        psixB[i] = 0;
        psicB[i] = 0;
    }
}

// Computes single step for psi_x: f(psixA, psicA, psixB, psicB, t).
// The results are returned in a freshly allocated array.  For spin σ
// = +1 call it with parameters (psixA, psicA, psixB, psicB, t), for
// spin σ = -1 call it with A and B parameters swapped: (psixA, psicA,
// psixB, psicB, t)
cdouble *compute_psixAB_substep(cdouble psixA[], cdouble psicA[],
                                cdouble psixB[], cdouble psicB[],
                                double t)
{
    cdouble *k = new cdouble[xsize];
    for (int i = 0; i < xsize; ++i) {
        k[i] = (-I * OmegaR * psicA[i] / hbar
                + ((-gammax * 0.5
                    - I * g1 * norm(psixA[i]) / hbar
                    - I * g2 * norm(psixB[i]) / hbar
                    - I * delta / hbar)
                   * psixA[i]));
    }
    return k;
}

// Computes single step for psi_c, spin σ = +1: f(psicA, psixA, t).
// The results are returned in a freshly allocated array.
cdouble *compute_psicA_substep(cdouble psicA[], cdouble psixA[], double t)
{
    cdouble *k = new cdouble[xsize];
    for (int i = 0; i < xsize; ++i) {
        const double x = i * sx + xmin;
        const cdouble psicm1 = i > 0 ? psicA[i - 1] : 0;
        const cdouble psicp1 = i < xsize - 1 ? psicA[i + 1] : 0;
        k[i] = (-I * FA(x, t) / hbar - I * OmegaR * psixA[i] / hbar
                - gammac * 0.5 * psicA[i]
                + (I * hbar / (2 * mc) / square(sx)
                   * (psicm1 - 2.0 * psicA[i] + psicp1)));
    }
    return k;
}

// Computes single step for psi_c, spin σ = -1: f(psicB, psixB, t).
// The results are returned in a freshly allocated array.
cdouble *compute_psicB_substep(cdouble psicB[], cdouble psixB[], double t)
{
    cdouble *k = new cdouble[xsize];
    for (int i = 0; i < xsize; ++i) {
        const double x = i * sx + xmin;
        const cdouble psicm1 = i > 0 ? psicB[i - 1] : 0;
        const cdouble psicp1 = i < xsize - 1 ? psicB[i + 1] : 0;
        k[i] = (-I * FB(x, t) / hbar - I * OmegaR * psixB[i] / hbar
                - gammac * 0.5 * psicB[i]
                + (I * hbar / (2 * mc) / square(sx)
                   * (psicm1 - 2.0 * psicB[i] + psicp1)));
    }
    return k;
}

// Computes x+y*c for scalar coefficient c and each pair of the
// elements of arrays x and y separately.  The results are returned in
// a freshly allocated array.
cdouble *increment_substep(cdouble x[], cdouble y[], double c)
{
    cdouble *k = new cdouble[xsize];
    for (int i = 0; i < xsize; ++i) {
        k[i] = x[i] + y[i] * c;
    }
    return k;
}

// Computes a single time step for psix and psic simultaneously.
void compute_step()
{
    // Substep 1.  Compute set of k1x and k1c coefficients starting
    // from psix and psic and time t for σ = ±1.
    cdouble *k1xA = compute_psixAB_substep(
        psixA, psicA, psixB, psicB, t);
    cdouble *k1cA = compute_psicA_substep(psicA, psixA, t);
    cdouble *k1xB = compute_psixAB_substep(
        psixB, psicB, psixA, psicA, t);
    cdouble *k1cB = compute_psicB_substep(psicB, psixB, t);

    // Substep 2a. Compute values of psix and psic for σ = ±1 at time
    // t + h / 2 using Euler's method.
    cdouble *psixA2 = increment_substep(psixA, k1xA, h * 0.5);
    cdouble *psicA2 = increment_substep(psicA, k1cA, h * 0.5);
    cdouble *psixB2 = increment_substep(psixB, k1xB, h * 0.5);
    cdouble *psicB2 = increment_substep(psicB, k1cB, h * 0.5);

    // Substep 2b. Compute set of k2x and k2c coefficients for σ = ±1.
    cdouble *k2xA = compute_psixAB_substep(
        psixA2, psicA2, psixB2, psicB2, t + h * 0.5);
    cdouble *k2cA = compute_psicA_substep(psicA2, psixA2, t + h * 0.5);
    cdouble *k2xB = compute_psixAB_substep(
        psixB2, psicB2, psixA2, psicA2, t + h * 0.5);
    cdouble *k2cB = compute_psicB_substep(psicB2, psixB2, t + h * 0.5);

    // Substep 3a. Compute values of psix and psic at time t + h / 2
    // using Euler's method for σ = ±1.
    cdouble *psixA3 = increment_substep(psixA, k2xA, h * 0.5);
    cdouble *psicA3 = increment_substep(psicA, k2cA, h * 0.5);
    cdouble *psixB3 = increment_substep(psixB, k2xB, h * 0.5);
    cdouble *psicB3 = increment_substep(psicB, k2cB, h * 0.5);

    // Substep 3b. Compute set of k3x and k3c coefficients for σ = ±1.
    cdouble *k3xA = compute_psixAB_substep(
        psixA3, psicA3, psixB3, psicB3, t + h * 0.5);
    cdouble *k3cA = compute_psicA_substep(psicA3, psixA3, t + h * 0.5);
    cdouble *k3xB = compute_psixAB_substep(
        psixB3, psicB3, psixA3, psicA3, t + h * 0.5);
    cdouble *k3cB = compute_psicB_substep(psicB3, psixB3, t + h * 0.5);

    // Substep 4a. Compute values of psix and psic at time t for σ = ±1.
    cdouble *psixA4 = increment_substep(psixA, k3xA, h);
    cdouble *psicA4 = increment_substep(psicA, k3cA, h);
    cdouble *psixB4 = increment_substep(psixB, k3xB, h);
    cdouble *psicB4 = increment_substep(psicB, k3cB, h);

    // Substep 4b. Compute set of k4x and k4c coefficients for σ = ±1.
    cdouble *k4xA = compute_psixAB_substep(
        psixA4, psicA4, psixB4, psicB4, t + h);
    cdouble *k4cA = compute_psicA_substep(psicA4, psixA4, t + h);
    cdouble *k4xB = compute_psixAB_substep(
        psixB4, psicB4, psixA4, psicA4, t + h);
    cdouble *k4cB = compute_psicB_substep(psicB4, psixB4, t + h);

    // Substep 5. Compute new values of psix and psic for t + h for σ = ±1.
    for (int i = 0; i < xsize; ++i) {
        psixA[i]
            += (k1xA[i] + 2.0 * k2xA[i] + 2.0 * k3xA[i] + k4xA[i]) * (h / 6);
        psicA[i]
            += (k1cA[i] + 2.0 * k2cA[i] + 2.0 * k3cA[i] + k4cA[i]) * (h / 6);
        psixB[i]
            += (k1xB[i] + 2.0 * k2xB[i] + 2.0 * k3xB[i] + k4xB[i]) * (h / 6);
        psicB[i]
            += (k1cB[i] + 2.0 * k2cB[i] + 2.0 * k3cB[i] + k4cB[i]) * (h / 6);
    }

    // Compute new t.
    t += h;

    // Free the memory.
    delete[] k1xA;
    delete[] k1cA;
    delete[] k2xA;
    delete[] k2cA;
    delete[] k3xA;
    delete[] k3cA;
    delete[] k4xA;
    delete[] k4cA;
    delete[] k1xB;
    delete[] k1cB;
    delete[] k2xB;
    delete[] k2cB;
    delete[] k3xB;
    delete[] k3cB;
    delete[] k4xB;
    delete[] k4cB;
    delete[] psixA2;
    delete[] psicA2;
    delete[] psixA3;
    delete[] psicA3;
    delete[] psixA4;
    delete[] psicA4;
    delete[] psixB2;
    delete[] psicB2;
    delete[] psixB3;
    delete[] psicB3;
    delete[] psixB4;
    delete[] psicB4;
}

int main(int argc, char* argv[])
{
    initialization();
    double maxv = 0;
    for (int n=0; n <= nsteps; ++n) {
        double sum = 0;
        for (int i = 0; i < xsize; ++i) {
            const double pxA = norm(psixA[i]);
            const double pcA = norm(psicA[i]);
            const double pxB = norm(psixB[i]);
            const double pcB = norm(psicB[i]);
            maxv = max(maxv, max(max(pxA, pcA), max(pxB, pcB)));
            sum += (pcA + pxA + pcB + pxB) * sx;
        }
        if (n % print_tstep == 0) {
            cerr << "t = " << t << ", sum = " << sum << endl;
            cout << setprecision(4) << t;
            for (int i = 0; i < xsize; ++i) {
                cout << '\t' << setprecision(5) << norm(psicA[i]);
            }
            for (int i = 0; i < xsize; ++i) {
                cout << '\t' << setprecision(5) << norm(psicB[i]);
            }
            for (int i = 0; i < xsize; ++i) {
                cout << '\t' << setprecision(5) << norm(psixA[i]);
            }
            for (int i = 0; i < xsize; ++i) {
                cout << '\t' << setprecision(5) << norm(psixB[i]);
            }
            cout << endl;
        }
        compute_step();
    }
    cerr << "maxv = " << maxv << endl;
}                             
