// EPCGP suite, 2D condensate, no spin effects
// ===========================================
//
// The set of two differential equations (Gross-Pitaevskii), with
// nonlinearities and with decay.  Program computes initial state and
// then uses it for solving the equation.  Every print_tstep it prints
// out the amplitudes of psi_x and psi_c and their Fourier transforms
// to four separate files.
//
// Parameters are set in the source code in order to ensure best
// optimization by the compiler.

#include <iostream>
#include <fstream>
#include <iomanip>
#include <cmath>
#include <complex>
#include <cstdlib>

using namespace std;

typedef complex<double> cdouble; 

// Helper constants.
const cdouble I = cdouble(0, 1); // Imaginary unit.
const double pi = 4 * atan(1);   // π.
const double hbar = 1;           // Reduced Planck constant [meV·ps].

// Parameters of the setup.
const double W = 2.0;           // Spread of the main Gaussian.
const double w = 0.25;          // Spread of the small Gaussian.
const double A = 0.5;           // Amplitude of the initial Gaussian.
const double k0 = 5;            // Must be greater than k_inflection.
const double OmegaR = 1;        // Rabi frequency [meV].
const double gammax = 0.01;     // Decay rate of an exciton [1/(ps)].
const double gammac = 0.01;     // Decay rate of a photon [1/(ps)].
const double delta = 0;         // Detuning of polaritons [meV].
const double g = 0.005;         // Material parameter [meV·m²].
const double mc = 1;            // Effective mass of polaritons [meV].
const double wb = 20;           // The width of the “safe” window [µm].
const double cavsizex = 24;     // Cavity size along X axis [µm].
const double cavsizey = 24;     // Cavity size along Y axis [µm].
const double h = 0.001;         // Time step [ps].

// File names.
const char *datxfname = "EPCGP_2D_psix.dat";
const char *datxFname = "EPCGP_2D_psixF.dat";
const char *datcfname = "EPCGP_2D_psic.dat";
const char *datcFname = "EPCGP_2D_psicF.dat";
const char *logname = "EPCGP_2D.log";

// Parameters of the simulation.
const int xsize = 241;          // Number of mesh nodes along X axis.
const int ysize = 241;          // Number of mesh nodes along Y axis.
const double sx = cavsizex / (xsize - 1); // Spatial step along X axis.
const double sy = cavsizey / (ysize - 1); // Spatial step along Y axis.
const double xmin = -(cavsizex / 2); // Lower boundary of X coordinate.
const double xmax = (cavsizex / 2); // Upper boundary of X coordinate.
const double ymin = -(cavsizey / 2); // Lower boundary of Y coordinate.
const double ymax = (cavsizey / 2); // Upper boundary of Y coordinate.
const double kxmin = -pi / sx;      // Lower boundary of k_x coordinate.
const double kxmax = pi / sx;       // Upper boundary of k_x coordinate.
const double kymin = -pi / sy;      // Lower boundary of k_y coordinate.
const double kymax = pi / sy;       // Upper boundary of k_y coordinate.
const double ksx = 2 * pi / cavsizex; // Frequency step along k_x axis.
const double ksy = 2 * pi / cavsizey; // Frequency step along k_y axis.
const double thsize = 2 * 360;        // Number of steps for angle theta.
const double thmin = 0;               // Minimal angle.
const double thmax = 2 * pi;          // Maximal angle.
const double sth = (thmax - thmin) / (thsize - 1); // Angle step.
const int nsteps = 2000;          // Number of simulation iterations.
const int print_tstep = 10;            // Print every nth time step.
const int print_xstep = 1;       // Print every nth spatial point along X axis.
const int print_ystep = 1;       // Print every nth spatial point along Y axis.

// Global variales.
cdouble psic[xsize * ysize];        // Array for values of psi_c [1/(µm)²].
cdouble psix[xsize * ysize];        // Array for values of psi_x [1/(µm)²].
double t;                       // Current time [ps].

// Computes square of complex x.
inline cdouble square(cdouble x)
{
    return x * x;
}

// Computes square of real x.
inline double square(double x) 
{
    return x * x;
}

// Computes discrete Fourier tranform of psi.  Returns psiF in a
// freshly allocated array.
cdouble *compute_Fourier(const cdouble *f)
{
    cdouble *r1 = new cdouble[xsize * ysize];
    cdouble *r2 = new cdouble[xsize * ysize];

    // 2D discrete Fourier transform is computed with two passes of 1D
    // Fourier transforms.
#pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        for (int j = 0; j < ysize; ++j) {
            const int q = i * ysize + j;
            const double ky = kymin + j * ksy;
            cdouble F = 0.;
            for (int jp = 0; jp < ysize; ++jp) {
                const int qp = i * ysize + jp;
                const double y = ymin + jp * sy;
                F += f[qp] * exp(-I * ky * y);
            }
            r1[q] = F / sqrt(ysize);
        }
    }
#pragma omp parallel for schedule(static)
    for (int j = 0; j < ysize; ++j) {
        for (int i = 0; i < xsize; ++i) {
            const int q = i * ysize + j;
            const double kx = kxmin + i * ksx;
            cdouble F = 0.;
            for (int ip = 0; ip < xsize; ++ip) {
                const int qp = ip * ysize + j;
                const double x = xmin + ip * sx;
                F += r1[qp] * exp(-I * kx * x);
            }
            r2[q] = F / sqrt(xsize);
        }
    }

    delete[] r1;
    return r2;
}

// Computes inverse discrete Fourier tranform of psiF.  Returns psi in
// a freshly allocated array.
cdouble *compute_InvFourier(const cdouble *F)
{
    cdouble *r1 = new cdouble[xsize * ysize];
    cdouble *r2 = new cdouble[xsize * ysize];

    // 2D discrete inverse Fourier transform is computed with two
    // passes of 1D Fourier transforms.
#pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        for (int j = 0; j < ysize; ++j) {
            const int q = i * ysize + j;
            const double y = ymin + j * sy;
            cdouble f = 0.;
            for (int jp = 0; jp < ysize; ++jp) {
                const int qp = i * ysize + jp;
                const double ky = kymin + jp * ksy;
                f += F[qp] * exp(I * ky * y);
            }
            r1[q] = f / sqrt(ysize);
        }
    }
#pragma omp parallel for schedule(static)
    for (int j = 0; j < ysize; ++j) {
        for (int i = 0; i < xsize; ++i) {
            const int q = i * ysize + j;
            const double x = xmin + i * sx;
            cdouble f = 0.;
            for (int ip = 0; ip < xsize; ++ip) {
                const int qp = ip * ysize + j;
                const double kx = kxmin + ip * ksx;
                f += r1[qp] * exp(I * kx * x);
            }
            r2[q] = f / sqrt(xsize);
        }
    }

    delete[] r1;
    return r2;
}

// Initializes psi and t.
void initialization()
{
    t = 0;

    cdouble *F0c = new cdouble[xsize * ysize];
    cdouble *F0x = new cdouble[xsize * ysize];

    // Preparation of F0c (the whole array is zeroed).
#pragma omp parallel for schedule(static)
    for (int j = 0; j < ysize; ++j) {
        for (int i = 0; i < xsize; ++i) {
            const int q = i * ysize + j;
            F0c[q] = 0;
        }
    }

    // First, F0c is "drawn".
    const double h22mc = square(hbar) / (2 * mc);
    const double epsk0 = h22mc * square(k0);
    const double a0 = h22mc / sqrt(square(epsk0) + 4 * square(OmegaR));
    bool new_point = true;
    double pkx = kxmin, pky = kymin;
    for (int t = 0; t <= thsize; ++t) {
        const double th = t * sth + thmin;
        const double k0sinth = k0 * sin(th);
        if (k0sinth == 0) {
            new_point = true;
            continue;
        }
        const double r
            = 1 / (2 * a0 * k0sinth) - (1 + a0 * square(k0)) * k0sinth;
        const double nkx = r * cos(th);
        const double nky = r * sin(th) + k0;
        if (new_point
            || pkx < kxmin || nkx < kxmin || pkx > kxmax || nkx > kxmax
            || pky < kymin || nky < kymin || pky > kymax || nky > kymax) {
            pkx = nkx;
            pky = nky;
            new_point = false;
            continue;
        }
        // The segment lies inside the area.  Let's draw it along the
        // longer axis.
        if (abs(nkx - pkx) >= abs(nky - pky)) {
            double startkx = pkx, endkx = nkx;
            double startky = pky, endky = nky;
            int mini = round((nkx - kxmin) / ksx);
            int maxi = round((pkx - kxmin) / ksx);
            if (mini > maxi) {
                int i = maxi;
                maxi = mini;
                mini = i;
                startkx = nkx; endkx = pkx;
                startky = nky; endky = pky;
            }
            mini = max(0, mini);
            maxi = min(maxi, xsize - 1);
            for (int i = mini; i <= maxi; ++i) {
                const double kx = kxmin + i * ksx;
                const double ky0
                    = (kx - startkx) / (endkx - startkx) * (endky - startky)
                    + startky;
                int minj = round((ky0 - 3 * w - kymin) / ksy);
                int maxj = round((ky0 + 3 * w - kymin) / ksy);
                if (minj > maxj) {
                    int j = maxj;
                    maxj = minj;
                    maxj = j;
                }
                minj = max(0, minj);
                maxj = min(maxj, ysize - 1);
                for (int j = minj; j <= maxj; ++j) {
                    const double ky = kymin + j * ksy;
                    const int q = i * ysize + j;
                    const cdouble F
                        = A * exp(-(square(ky - k0) + square(kx))
                                  / (2 * square(W)))
                        / sqrt(2 * pi * square(w))
                        * exp(-square(ky - ky0) / (2 * square(w)));
                    F0c[q] = abs(F0c[q]) > abs(F) ? F0c[q] : F;
                }
            }
        } else {
            double startkx = pkx, endkx = nkx;
            double startky = pky, endky = nky;
            int minj = round((nky - kymin) / ksy);
            int maxj = round((pky - kymin) / ksy);
            if (minj > maxj) {
                int j = maxj;
                maxj = minj;
                minj = j;
                startkx = nkx; endkx = pkx;
                startky = nky; endky = pky;
            }
            minj = max(0, minj);
            maxj = min(maxj, ysize - 1);
            for (int j = minj; j <= maxj; ++j) {
                const double ky = kymin + j * ksy;
                const double kx0
                    = (ky - startky) / (endky - startky) * (endkx - startkx)
                    + startkx;
                int mini = round((kx0 - 3 * w - kxmin) / ksx);
                int maxi = round((kx0 + 3 * w - kxmin) / ksx);
                if (mini > maxi) {
                    int i = maxi;
                    maxi = mini;
                    maxi = i;
                }
                mini = max(0, mini);
                maxi = min(maxi, xsize - 1);
                for (int i = mini; i <= maxi; ++i) {
                    const double kx = kxmin + i * ksx;
                    const int q = i * ysize + j;
                    const cdouble F
                        = A * exp(-(square(ky - k0) + square(kx))
                                  / (2 * square(W)))
                        / sqrt(2 * pi * square(w))
                        * exp(-square(kx - kx0) / (2 * square(w)));
                    F0c[q] = abs(F0c[q]) > abs(F) ? F0c[q] : F;
                }
            }
        }
        pkx = nkx;
        pky = nky;
    }

    // F0x is computed from F0c.
#pragma omp parallel for schedule(static)
    for (int j = 0; j < ysize; ++j) {
        const double ky = j * ksy + kymin;
        for (int i = 0; i < xsize; ++i) {
            const double kx = i * ksx + kxmin;
            const double epsk = square(hbar) / (2 * mc) * (
                square(kx) + square(ky));
            const double hbaromega = 0.5 * (
                epsk - sqrt(4 * square(OmegaR) + square(epsk)));
            const int q = i * ysize + j;
            F0x[q] = OmegaR / hbaromega * F0c[q];
        }
    }

    // Computation of the initial psic and psix state as an inverse of
    // F0c and F0x.
    cdouble *psi0c = compute_InvFourier(F0c);
    cdouble *psi0x = compute_InvFourier(F0x);

    // Save psi0c to psic and and psi0x to psix.
#pragma omp parallel for schedule(static)
    for (int i = 0; i < (xsize * ysize); ++i) {
        psic[i] = psi0c[i];
        psix[i] = psi0x[i];
    }

    delete[] F0c; delete[] F0x;
    delete[] psi0c; delete[] psi0x;
}

// Computes single step for psi: f(psi, t).  The results are returned
// in a freshly allocated array.
cdouble *compute_psic_substep(cdouble psic[], cdouble psix[], double t)
{
    cdouble *r = new cdouble[xsize * ysize];
#pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        for (int j = 0; j < ysize; ++j) {
            const int q = i * ysize + j;
            const cdouble psicxm1 = i > 0 ? psic[(i - 1) * ysize + j] : 0;
            const cdouble psicxp1
                = i < xsize - 1 ? psic[(i + 1) * ysize + j] : 0;
            const cdouble psicym1
                = j > 0 ? psic[q - 1] : psic[i * ysize + ysize - 1];
            const cdouble psicyp1
                = j < ysize - 1 ? psic[q + 1] : psic[i * ysize];
            r[q] = (-I * OmegaR / hbar * psix[q]
                    + (I * hbar / (2 * mc) / square(sx)
                       * (psicxm1 - 2.0 * psic[q] + psicxp1))
                    + (I * hbar / (2 * mc) / square(sy)
                       * (psicym1 - 2.0 * psic[q] + psicyp1))
                    - (gammac / 2 * psic[q]));
        }
    }
    return r;
}

// Computes single step for psi: f(psi, t).  The results are returned
// in a freshly allocated array.
cdouble *compute_psix_substep(cdouble psix[], cdouble psic[], double t)
{
    cdouble *r = new cdouble[xsize * ysize];
#pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        for (int j = 0; j < ysize; ++j) {
            const int q = i * ysize + j;
            r[q] = (-I * OmegaR / hbar * psic[q]
                    - (gammax / 2 * psix[q])
                    - (I * g / hbar * square(abs(psix[q])) * psix[q]));
        }
    }
    return r;
}

// Computes x+y*c for scalar coefficient c and each pair of the
// elements of arrays x and y separately.  The results are returned in
// a freshly allocated array.
cdouble *increment_substep(cdouble x[], cdouble y[], double c)
{
    cdouble *r = new cdouble[xsize * ysize];
#pragma omp parallel for schedule(static)
    for (int i = 0; i < (xsize * ysize); ++i) {
        r[i] = x[i] + y[i] * c;
    }
    return r;
}

// Computes a single time step for psix and psic simultaneously.
void compute_step()
{
    // Substep 1.  Compute coefficients k1.
    cdouble *k1c = compute_psic_substep(psic, psix, t);
    cdouble *k1x = compute_psix_substep(psix, psic, t);

    // Substep 2a. Compute values of psi at time t + h / 2 using
    // Euler's method.
    cdouble *psi2c = increment_substep(psic, k1c, h * 0.5);
    cdouble *psi2x = increment_substep(psix, k1x, h * 0.5);

    // Substep 2b. Compute coefficients k2.
    cdouble *k2c = compute_psic_substep(psi2c, psi2x, t + h * 0.5);
    cdouble *k2x = compute_psix_substep(psi2x, psi2c, t + h * 0.5);

    // Substep 3a. Compute values of psi at time t + h / 2 using
    // Euler's method.
    cdouble *psi3c = increment_substep(psic, k2c, h * 0.5);
    cdouble *psi3x = increment_substep(psix, k2x, h * 0.5);

    // Substep 3b. Compute coefficients k3.
    cdouble *k3c = compute_psic_substep(psi3c, psi3x, t + h * 0.5);
    cdouble *k3x = compute_psix_substep(psi3x, psi3c, t + h * 0.5);

    // Substep 4a. Compute values of psi at time t.
    cdouble *psi4c = increment_substep(psic, k3c, h);
    cdouble *psi4x = increment_substep(psix, k3x, h);

    // Substep 4b. Compute coefficients k4.
    cdouble *k4c = compute_psic_substep(psi4c, psi4x, t + h);
    cdouble *k4x = compute_psix_substep(psi4x, psi4c, t + h);

    // Substep 5. Compute new values of psi for t + h.
#pragma omp parallel for schedule(static)
    for (int i = 0; i < (xsize * ysize); ++i) {
        psic[i] += (k1c[i] + 2.0 * k2c[i] + 2.0 * k3c[i] + k4c[i]) * (h / 6);
        psix[i] += (k1x[i] + 2.0 * k2x[i] + 2.0 * k3x[i] + k4x[i]) * (h / 6);
    }

    // Compute new t.
    t += h;

    // Free the memory.
    delete[] k1c; delete[] k2c; delete[] k3c; delete[] k4c;
    delete[] k1x; delete[] k2x; delete[] k3x; delete[] k4x;
    delete[] psi2c; delete[] psi3c; delete[] psi4c;
    delete[] psi2x; delete[] psi3x; delete[] psi4x;
}

// Applies a safety window at the mesh edges.
void apply_safety_window()
{
#pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        const double x = i * sx + xmin;
        const double cx = exp(-pow(x / wb, 20));
        for (int j = 0; j < ysize; ++j) {
            const int q = i * ysize + j;
            psic[q] *= cx;
            psix[q] *= cx;
        }
    }
}

// Helper function.  Sends the given array of squared values to the
// given file.
void print_array(cdouble x[], ofstream &of)
{
    for (int i = 0; i < xsize; i += print_xstep) {
        for (int j = 0; j < ysize; j += print_ystep) {
            if (j > 0)
                of << '\t';
            of << setprecision(5) << square(abs(x[i * ysize + j]));
        }
        of << endl;
    }
}

int main(int argc, char* argv[])
{
    ofstream fpsic(datcfname);
    ofstream fpsicF(datcFname);
    ofstream fpsix(datxfname);
    ofstream fpsixF(datxFname);
    ofstream flog(logname);

    initialization();
    apply_safety_window();
    double maxvc = 0, maxvx = 0, maxvcF = 0, maxvxF = 0;
    for (int n=0; n <= nsteps; ++n) {
        double sumc = 0, sumx = 0;
        for (int i = 0; i < (xsize * ysize); ++i) {
            const double pc = square(abs(psic[i]));
            const double px = square(abs(psix[i]));
            maxvc = max(maxvc, pc);
            maxvx = max(maxvx, px);
            sumc += pc * sx * sy;
            sumx += px * sx * sy;
        }
        if (n % print_tstep == 0) {
            print_array(psic, fpsic);
            print_array(psix, fpsix);
            cdouble *psicF = compute_Fourier(psic);
            cdouble *psixF = compute_Fourier(psix);
            print_array(psicF, fpsicF);
            print_array(psixF, fpsixF);
            double sumcF = 0, sumxF = 0;
            for (int i = 0; i < (xsize * ysize); ++i) {
                const double pcF = square(abs(psicF[i]));
                const double pxF = square(abs(psixF[i]));
                maxvcF = max(maxvcF, pcF);
                maxvxF = max(maxvxF, pxF);
                sumcF += pcF * ksx * ksy;
                sumxF += pxF * ksx * ksy;
            }
            cerr << "t = " << t
                 << ", sumc = " << sumc
                 << ", sumx = " << sumx
                 << ", sumcF = " << sumcF
                 << ", sumxF = " << sumxF << endl;
            flog << "t = " << t
                 << ", sumc = " << sumc
                 << ", sumx = " << sumx
                 << ", sumcF = " << sumcF
                 << ", sumxF = " << sumxF << endl;
            delete[] psicF; delete[] psixF;
        }
        compute_step();
        apply_safety_window();
    }
    cerr << "maxvc = " << maxvc
         << ", maxvx = " << maxvx
         << ", maxvcF = " << maxvcF
         << ", maxvxF = " << maxvxF << endl;
    flog << "maxvc = " << maxvc
         << ", maxvx = " << maxvx
         << ", maxvcF = " << maxvcF
         << ", maxvxF = " << maxvxF << endl;
    fpsic.close();
    fpsicF.close();
    fpsix.close();
    fpsixF.close();
    flog.close();
}                             
