// EPCGP suite, 2D condensate, spin effects included
// =================================================
//
// The set of four differential equations (Gross-Pitaevskii), with
// nonlinearities and with decay.  Program computes initial state and
// then uses it for solving the equation.  Every print_tstep it prints
// out the amplitudes of psi_x and psi_c and their Fourier transforms
// to eight separate files.
//
// Parameters are set in the source code in order to ensure best
// optimization by the compiler.

#include <iostream>
#include <fstream>
#include <iomanip>
#include <cmath>
#include <complex>
#include <cstdlib>

using namespace std;

typedef complex<double> cdouble; 

// Helper constants.
const cdouble I = cdouble(0, 1); // Imaginary unit.
const double pi = 4 * atan(1);   // π.
const double hbar = 1;           // Reduced Planck constant [meV·ps].

// Parameters of the setup.
const double W = 2.0;           // Spread of the main Gaussian.
const double w = 0.25;          // Spread of the small Gaussian.
const double A = 0.5;           // Amplitude of the initial Gaussian.
const double k0 = 5;            // Must be greater than k_inflection.
const double OmegaR = 1;        // Rabi frequency [meV].
const double gammax = 0.01;     // Decay rate of an exciton [1/(ps)].
const double gammac = 0.01;     // Decay rate of a photon [1/(ps)].
const double delta = 0;         // Detuning of polaritons [meV].
const double g1 = 0.005;        // Material parameter [meV·m²].
const double g2 = 0.005;        // Material parameter [meV·m²].
const double mc = 1;            // Effective mass of polaritons [meV].
const double wb = 20;           // The width of the “safe” window [µm].
const double cavsizex = 24;     // Cavity size along X axis [µm].
const double cavsizey = 24;     // Cavity size along Y axis [µm].
const double h = 0.0001;        // Time step [ps].

// File names.
const char *datxfnameA = "EPCGP_2D_spin_psixA.dat";
const char *datxFnameA = "EPCGP_2D_spin_psixFA.dat";
const char *datcfnameA = "EPCGP_2D_spin_psicA.dat";
const char *datcFnameA = "EPCGP_2D_spin_psicFA.dat";
const char *datxfnameB = "EPCGP_2D_spin_psixB.dat";
const char *datxFnameB = "EPCGP_2D_spin_psixFB.dat";
const char *datcfnameB = "EPCGP_2D_spin_psicB.dat";
const char *datcFnameB = "EPCGP_2D_spin_psicFB.dat";
const char *logname = "EPCGP_2D_spin.log";

// Parameters of the simulation.
const int xsize = 241;          // Number of mesh nodes along X axis.
const int ysize = 241;          // Number of mesh nodes along Y axis.
const double sx = cavsizex / (xsize - 1); // Spatial step along X axis.
const double sy = cavsizey / (ysize - 1); // Spatial step along Y axis.
const double xmin = -(cavsizex / 2); // Lower boundary of X coordinate.
const double xmax = (cavsizex / 2); // Upper boundary of X coordinate.
const double ymin = -(cavsizey / 2); // Lower boundary of Y coordinate.
const double ymax = (cavsizey / 2); // Upper boundary of Y coordinate.
const double kxmin = -pi / sx;      // Lower boundary of k_x coordinate.
const double kxmax = pi / sx;       // Upper boundary of k_x coordinate.
const double kymin = -pi / sy;      // Lower boundary of k_y coordinate.
const double kymax = pi / sy;       // Upper boundary of k_y coordinate.
const double ksx = 2 * pi / cavsizex; // Frequency step along k_x axis.
const double ksy = 2 * pi / cavsizey; // Frequency step along k_y axis.
const double thsize = 2 * 360;        // Number of steps for angle theta.
const double thmin = 0;               // Minimal angle.
const double thmax = 2 * pi;          // Maximal angle.
const double sth = (thmax - thmin) / (thsize - 1); // Angle step.
const int nsteps = 2000;          // Number of simulation iterations.
const int print_tstep = 10;            // Print every nth time step.
const int print_xstep = 1;       // Print every nth spatial point along X axis.
const int print_ystep = 1;       // Print every nth spatial point along Y axis.

// Global variales.
cdouble psicA[xsize * ysize];// Array for values of psi_c, spin σ = +1 [1/(µm)²].
cdouble psixA[xsize * ysize];// Array for values of psi_x, spin σ = +1 [1/(µm)²].
cdouble psicB[xsize * ysize];// Array for values of psi_c, spin σ = -1 [1/(µm)²].
cdouble psixB[xsize * ysize];// Array for values of psi_x, spin σ = -1 [1/(µm)²].
double t;                    // Current time [ps].

// Computes square of complex x.
inline cdouble square(cdouble x)
{
    return x * x;
}

// Computes square of real x.
inline double square(double x) 
{
    return x * x;
}

// Computes discrete Fourier tranform of psi.  Returns psiF in a
// freshly allocated array.
cdouble *compute_Fourier(const cdouble *f)
{
    cdouble *r1 = new cdouble[xsize * ysize];
    cdouble *r2 = new cdouble[xsize * ysize];

    // 2D discrete Fourier transform is computed with two passes of 1D
    // Fourier transforms.
#pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        for (int j = 0; j < ysize; ++j) {
            const int q = i * ysize + j;
            const double ky = kymin + j * ksy;
            cdouble F = 0.;
            for (int jp = 0; jp < ysize; ++jp) {
                const int qp = i * ysize + jp;
                const double y = ymin + jp * sy;
                F += f[qp] * exp(-I * ky * y);
            }
            r1[q] = F / sqrt(ysize);
        }
    }
#pragma omp parallel for schedule(static)
    for (int j = 0; j < ysize; ++j) {
        for (int i = 0; i < xsize; ++i) {
            const int q = i * ysize + j;
            const double kx = kxmin + i * ksx;
            cdouble F = 0.;
            for (int ip = 0; ip < xsize; ++ip) {
                const int qp = ip * ysize + j;
                const double x = xmin + ip * sx;
                F += r1[qp] * exp(-I * kx * x);
            }
            r2[q] = F / sqrt(xsize);
        }
    }

    delete[] r1;
    return r2;
}

// Computes inverse discrete Fourier tranform of psiF.  Returns psi in
// a freshly allocated array.
cdouble *compute_InvFourier(const cdouble *F)
{
    cdouble *r1 = new cdouble[xsize * ysize];
    cdouble *r2 = new cdouble[xsize * ysize];

    // 2D discrete inverse Fourier transform is computed with two
    // passes of 1D Fourier transforms.
#pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        for (int j = 0; j < ysize; ++j) {
            const int q = i * ysize + j;
            const double y = ymin + j * sy;
            cdouble f = 0.;
            for (int jp = 0; jp < ysize; ++jp) {
                const int qp = i * ysize + jp;
                const double ky = kymin + jp * ksy;
                f += F[qp] * exp(I * ky * y);
            }
            r1[q] = f / sqrt(ysize);
        }
    }
#pragma omp parallel for schedule(static)
    for (int j = 0; j < ysize; ++j) {
        for (int i = 0; i < xsize; ++i) {
            const int q = i * ysize + j;
            const double x = xmin + i * sx;
            cdouble f = 0.;
            for (int ip = 0; ip < xsize; ++ip) {
                const int qp = ip * ysize + j;
                const double kx = kxmin + ip * ksx;
                f += r1[qp] * exp(I * kx * x);
            }
            r2[q] = f / sqrt(xsize);
        }
    }

    delete[] r1;
    return r2;
}

// Initializes psi and t.
void initialization()
{
    t = 0;

    cdouble *F0c = new cdouble[xsize * ysize];
    cdouble *F0x = new cdouble[xsize * ysize];

    // Preparation of F0c (the whole array is zeroed).
#pragma omp parallel for schedule(static)
    for (int j = 0; j < ysize; ++j) {
        for (int i = 0; i < xsize; ++i) {
            const int q = i * ysize + j;
            F0c[q] = 0;
        }
    }

    // First, F0c is "drawn".
    const double h22mc = square(hbar) / (2 * mc);
    const double epsk0 = h22mc * square(k0);
    const double a0 = h22mc / sqrt(square(epsk0) + 4 * square(OmegaR));
    bool new_point = true;
    double pkx = kxmin, pky = kymin;
    for (int t = 0; t <= thsize; ++t) {
        const double th = t * sth + thmin;
        const double k0sinth = k0 * sin(th);
        if (k0sinth == 0) {
            new_point = true;
            continue;
        }
        const double r
            = 1 / (2 * a0 * k0sinth) - (1 + a0 * square(k0)) * k0sinth;
        const double nkx = r * cos(th);
        const double nky = r * sin(th) + k0;
        if (new_point
            || pkx < kxmin || nkx < kxmin || pkx > kxmax || nkx > kxmax
            || pky < kymin || nky < kymin || pky > kymax || nky > kymax) {
            pkx = nkx;
            pky = nky;
            new_point = false;
            continue;
        }
        // The segment lies inside the area.  Let's draw it along the
        // longer axis.
        if (abs(nkx - pkx) >= abs(nky - pky)) {
            double startkx = pkx, endkx = nkx;
            double startky = pky, endky = nky;
            int mini = round((nkx - kxmin) / ksx);
            int maxi = round((pkx - kxmin) / ksx);
            if (mini > maxi) {
                int i = maxi;
                maxi = mini;
                mini = i;
                startkx = nkx; endkx = pkx;
                startky = nky; endky = pky;
            }
            mini = max(0, mini);
            maxi = min(maxi, xsize - 1);
            for (int i = mini; i <= maxi; ++i) {
                const double kx = kxmin + i * ksx;
                const double ky0
                    = (kx - startkx) / (endkx - startkx) * (endky - startky)
                    + startky;
                int minj = round((ky0 - 3 * w - kymin) / ksy);
                int maxj = round((ky0 + 3 * w - kymin) / ksy);
                if (minj > maxj) {
                    int j = maxj;
                    maxj = minj;
                    maxj = j;
                }
                minj = max(0, minj);
                maxj = min(maxj, ysize - 1);
                for (int j = minj; j <= maxj; ++j) {
                    const double ky = kymin + j * ksy;
                    const int q = i * ysize + j;
                    const cdouble F
                        = A * exp(-(square(ky - k0) + square(kx))
                                  / (2 * square(W)))
                        / sqrt(2 * pi * square(w))
                        * exp(-square(ky - ky0) / (2 * square(w)));
                    F0c[q] = abs(F0c[q]) > abs(F) ? F0c[q] : F;
                }
            }
        } else {
            double startkx = pkx, endkx = nkx;
            double startky = pky, endky = nky;
            int minj = round((nky - kymin) / ksy);
            int maxj = round((pky - kymin) / ksy);
            if (minj > maxj) {
                int j = maxj;
                maxj = minj;
                minj = j;
                startkx = nkx; endkx = pkx;
                startky = nky; endky = pky;
            }
            minj = max(0, minj);
            maxj = min(maxj, ysize - 1);
            for (int j = minj; j <= maxj; ++j) {
                const double ky = kymin + j * ksy;
                const double kx0
                    = (ky - startky) / (endky - startky) * (endkx - startkx)
                    + startkx;
                int mini = round((kx0 - 3 * w - kxmin) / ksx);
                int maxi = round((kx0 + 3 * w - kxmin) / ksx);
                if (mini > maxi) {
                    int i = maxi;
                    maxi = mini;
                    maxi = i;
                }
                mini = max(0, mini);
                maxi = min(maxi, xsize - 1);
                for (int i = mini; i <= maxi; ++i) {
                    const double kx = kxmin + i * ksx;
                    const int q = i * ysize + j;
                    const cdouble F
                        = A * exp(-(square(ky - k0) + square(kx))
                                  / (2 * square(W)))
                        / sqrt(2 * pi * square(w))
                        * exp(-square(kx - kx0) / (2 * square(w)));
                    F0c[q] = abs(F0c[q]) > abs(F) ? F0c[q] : F;
                }
            }
        }
        pkx = nkx;
        pky = nky;
    }

    // F0x is computed from F0c.
#pragma omp parallel for schedule(static)
    for (int j = 0; j < ysize; ++j) {
        const double ky = j * ksy + kymin;
        for (int i = 0; i < xsize; ++i) {
            const double kx = i * ksx + kxmin;
            const double epsk = square(hbar) / (2 * mc) * (
                square(kx) + square(ky));
            const double hbaromega = 0.5 * (
                epsk - sqrt(4 * square(OmegaR) + square(epsk)));
            const int q = i * ysize + j;
            F0x[q] = OmegaR / hbaromega * F0c[q];
        }
    }

    // Computation of the initial psic and psix state as an inverse of
    // F0c and F0x.
    cdouble *psi0c = compute_InvFourier(F0c);
    cdouble *psi0x = compute_InvFourier(F0x);

    // Save psi0c to psic and and psi0x to psix.
#pragma omp parallel for schedule(static)
    for (int i = 0; i < (xsize * ysize); ++i) {
        psicA[i] = psi0c[i];
        psixA[i] = psi0x[i];
        psicB[i] = psi0c[i];
        psixB[i] = psi0x[i];
    }

    delete[] F0c; delete[] F0x;
    delete[] psi0c; delete[] psi0x;
}

// Computes single step for psi: f(psi, t).  The results are returned
// in a freshly allocated array.
cdouble *compute_psic_substep(cdouble psic[], cdouble psix[], double t)
{
    cdouble *r = new cdouble[xsize * ysize];
#pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        for (int j = 0; j < ysize; ++j) {
            const int q = i * ysize + j;
            const cdouble psicxm1 = i > 0 ? psic[(i - 1) * ysize + j] : 0;
            const cdouble psicxp1
                = i < xsize - 1 ? psic[(i + 1) * ysize + j] : 0;
            const cdouble psicym1
                = j > 0 ? psic[q - 1] : psic[i * ysize + ysize - 1];
            const cdouble psicyp1
                = j < ysize - 1 ? psic[q + 1] : psic[i * ysize];
            r[q] = (-I * OmegaR / hbar * psix[q]
                    + (I * hbar / (2 * mc) / square(sx)
                       * (psicxm1 - 2.0 * psic[q] + psicxp1))
                    + (I * hbar / (2 * mc) / square(sy)
                       * (psicym1 - 2.0 * psic[q] + psicyp1))
                    - (gammac / 2 * psic[q]));
        }
    }
    return r;
}

// Computes single step for psi: f(psi, t).  The results are returned
// in a freshly allocated array.
cdouble *compute_psix_substep(cdouble psix[], cdouble psic[], cdouble psix2[], double t)
{
    cdouble *r = new cdouble[xsize * ysize];
#pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        for (int j = 0; j < ysize; ++j) {
            const int q = i * ysize + j;
            r[q] = (-I * OmegaR / hbar * psic[q]
                    - (gammax / 2 * psix[q])
                    - (I * g1 / hbar * square(abs(psix[q])) * psix[q])
                    - (I * g2 / hbar * square(abs(psix2[q])) * psix[q]));
        }
    }
    return r;
}

// Computes x+y*c for scalar coefficient c and each pair of the
// elements of arrays x and y separately.  The results are returned in
// a freshly allocated array.
cdouble *increment_substep(cdouble x[], cdouble y[], double c)
{
    cdouble *r = new cdouble[xsize * ysize];
#pragma omp parallel for schedule(static)
    for (int i = 0; i < (xsize * ysize); ++i) {
        r[i] = x[i] + y[i] * c;
    }
    return r;
}

// Computes a single time step for psix and psic simultaneously.
void compute_step()
{
    // Substep 1.  Compute coefficients k1.
    cdouble *k1cA = compute_psic_substep(psicA, psixA, t);
    cdouble *k1xA = compute_psix_substep(psixA, psicA, psixB, t);
    cdouble *k1cB = compute_psic_substep(psicB, psixB, t);
    cdouble *k1xB = compute_psix_substep(psixB, psicB, psixA, t);

    // Substep 2a. Compute values of psi at time t + h / 2 using
    // Euler's method.
    cdouble *psi2cA = increment_substep(psicA, k1cA, h * 0.5);
    cdouble *psi2xA = increment_substep(psixA, k1xA, h * 0.5);
    cdouble *psi2cB = increment_substep(psicB, k1cB, h * 0.5);
    cdouble *psi2xB = increment_substep(psixB, k1xB, h * 0.5);

    // Substep 2b. Compute coefficients k2.
    cdouble *k2cA = compute_psic_substep(psi2cA, psi2xA, t + h * 0.5);
    cdouble *k2xA = compute_psix_substep(psi2xA, psi2cA, psi2xB, t + h * 0.5);
    cdouble *k2cB = compute_psic_substep(psi2cB, psi2xB, t + h * 0.5);
    cdouble *k2xB = compute_psix_substep(psi2xB, psi2cB, psi2xA, t + h * 0.5);

    // Substep 3a. Compute values of psi at time t + h / 2 using
    // Euler's method.
    cdouble *psi3cA = increment_substep(psicA, k2cA, h * 0.5);
    cdouble *psi3xA = increment_substep(psixA, k2xA, h * 0.5);
    cdouble *psi3cB = increment_substep(psicB, k2cB, h * 0.5);
    cdouble *psi3xB = increment_substep(psixB, k2xB, h * 0.5);

    // Substep 3b. Compute coefficients k3.
    cdouble *k3cA = compute_psic_substep(psi3cA, psi3xA, t + h * 0.5);
    cdouble *k3xA = compute_psix_substep(psi3xA, psi3cA, psi3xB, t + h * 0.5);
    cdouble *k3cB = compute_psic_substep(psi3cB, psi3xB, t + h * 0.5);
    cdouble *k3xB = compute_psix_substep(psi3xB, psi3cB, psi3xA, t + h * 0.5);

    // Substep 4a. Compute values of psi at time t.
    cdouble *psi4cA = increment_substep(psicA, k3cA, h);
    cdouble *psi4xA = increment_substep(psixA, k3xA, h);
    cdouble *psi4cB = increment_substep(psicB, k3cB, h);
    cdouble *psi4xB = increment_substep(psixB, k3xB, h);

    // Substep 4b. Compute coefficients k4.
    cdouble *k4cA = compute_psic_substep(psi4cA, psi4xA, t + h);
    cdouble *k4xA = compute_psix_substep(psi4xA, psi4cA, psi4xB, t + h);
    cdouble *k4cB = compute_psic_substep(psi4cB, psi4xB, t + h);
    cdouble *k4xB = compute_psix_substep(psi4xB, psi4cB, psi4xA, t + h);

    // Substep 5. Compute new values of psi for t + h.
#pragma omp parallel for schedule(static)
    for (int i = 0; i < (xsize * ysize); ++i) {
        psicA[i] += (k1cA[i] + 2.0 * k2cA[i] + 2.0 * k3cA[i] + k4cA[i]) * (h / 6);
        psixA[i] += (k1xA[i] + 2.0 * k2xA[i] + 2.0 * k3xA[i] + k4xA[i]) * (h / 6);
        psicB[i] += (k1cB[i] + 2.0 * k2cB[i] + 2.0 * k3cB[i] + k4cB[i]) * (h / 6);
        psixB[i] += (k1xB[i] + 2.0 * k2xB[i] + 2.0 * k3xB[i] + k4xB[i]) * (h / 6);
    }

    // Compute new t.
    t += h;

    // Free the memory.
    delete[] k1cA; delete[] k2cA; delete[] k3cA; delete[] k4cA;
    delete[] k1xA; delete[] k2xA; delete[] k3xA; delete[] k4xA;
    delete[] k1cB; delete[] k2cB; delete[] k3cB; delete[] k4cB;
    delete[] k1xB; delete[] k2xB; delete[] k3xB; delete[] k4xB;
    delete[] psi2cA; delete[] psi3cA; delete[] psi4cA;
    delete[] psi2xA; delete[] psi3xA; delete[] psi4xA;
    delete[] psi2cB; delete[] psi3cB; delete[] psi4cB;
    delete[] psi2xB; delete[] psi3xB; delete[] psi4xB;
}

// Applies a safety window at the mesh edges.
void apply_safety_window()
{
#pragma omp parallel for schedule(static)
    for (int i = 0; i < xsize; ++i) {
        const double x = i * sx + xmin;
        const double cx = exp(-pow(x / wb, 20));
        for (int j = 0; j < ysize; ++j) {
            const int q = i * ysize + j;
            psicA[q] *= cx;
            psixA[q] *= cx;
            psicB[q] *= cx;
            psixB[q] *= cx;
        }
    }
}

// Helper function.  Sends the given array of squared values to the
// given file.
void print_array(cdouble x[], ofstream &of)
{
    for (int i = 0; i < xsize; i += print_xstep) {
        for (int j = 0; j < ysize; j += print_ystep) {
            if (j > 0)
                of << '\t';
            of << setprecision(5) << square(abs(x[i * ysize + j]));
        }
        of << endl;
    }
}

int main(int argc, char* argv[])
{
    ofstream fpsicA(datcfnameA);
    ofstream fpsicFA(datcFnameA);
    ofstream fpsixA(datxfnameA);
    ofstream fpsixFA(datxFnameA);
    ofstream fpsicB(datcfnameB);
    ofstream fpsicFB(datcFnameB);
    ofstream fpsixB(datxfnameB);
    ofstream fpsixFB(datxFnameB);
    ofstream flog(logname);

    initialization();
    apply_safety_window();
    double maxvc = 0, maxvx = 0, maxvcF = 0, maxvxF = 0;
    for (int n=0; n <= nsteps; ++n) {
        double sumc = 0, sumx = 0;
        for (int i = 0; i < (xsize * ysize); ++i) {
            const double pc = square(abs(psicA[i]));
            const double px = square(abs(psixA[i]));
            maxvc = max(maxvc, pc);
            maxvx = max(maxvx, px);
            sumc += pc * sx * sy;
            sumx += px * sx * sy;
        }
        if (n % print_tstep == 0) {
            print_array(psicA, fpsicA);
            print_array(psixA, fpsixA);
            print_array(psicB, fpsicB);
            print_array(psixB, fpsixB);
            cdouble *psicFA = compute_Fourier(psicA);
            cdouble *psixFA = compute_Fourier(psixA);
            cdouble *psicFB = compute_Fourier(psicB);
            cdouble *psixFB = compute_Fourier(psixB);
            print_array(psicFA, fpsicFA);
            print_array(psixFA, fpsixFA);
            print_array(psicFB, fpsicFB);
            print_array(psixFB, fpsixFB);
            double sumcF = 0, sumxF = 0;
            for (int i = 0; i < (xsize * ysize); ++i) {
                const double pcF = square(abs(psicFA[i]));
                const double pxF = square(abs(psixFA[i]));
                maxvcF = max(maxvcF, pcF);
                maxvxF = max(maxvxF, pxF);
                sumcF += pcF * ksx * ksy;
                sumxF += pxF * ksx * ksy;
            }
            cerr << "t = " << t
                 << ", sumc = " << sumc
                 << ", sumx = " << sumx
                 << ", sumcF = " << sumcF
                 << ", sumxF = " << sumxF << endl;
            flog << "t = " << t
                 << ", sumc = " << sumc
                 << ", sumx = " << sumx
                 << ", sumcF = " << sumcF
                 << ", sumxF = " << sumxF << endl;
            delete[] psicFA; delete[] psixFA;
            delete[] psicFB; delete[] psixFB;
        }
        compute_step();
        apply_safety_window();
    }
    cerr << "maxvc = " << maxvc
         << ", maxvx = " << maxvx
         << ", maxvcF = " << maxvcF
         << ", maxvxF = " << maxvxF << endl;
    flog << "maxvc = " << maxvc
         << ", maxvx = " << maxvx
         << ", maxvcF = " << maxvcF
         << ", maxvxF = " << maxvxF << endl;
    fpsicA.close();
    fpsicFA.close();
    fpsixA.close();
    fpsixFA.close();
    fpsicB.close();
    fpsicFB.close();
    fpsixB.close();
    fpsixFB.close();
    flog.close();
}                             
